using System;
using System.Text;
using System.IO;
using System.Threading;

namespace AsyncFileRead
{
	class FileReaderDemo
	{
		const string FILE_NAME = "data.txt";
		const int READ_BUF_SIZE = 8192;
		const int WAIT_TIMEOUT = 50;

		public void CreateFile()
		{
			StreamWriter writer = File.CreateText(FILE_NAME);
			using (writer)
			{
				writer.Write("12345678901234567890");
			}
		}

		public void SynchronousReadFile()
		{
			Stream stream = File.OpenRead(FILE_NAME);
			using (stream)
			{
				byte[] buf = new byte[READ_BUF_SIZE];
				int bytesRead = stream.Read(buf, 0, buf.Length);
				string data = Encoding.ASCII.GetString(buf, 0, bytesRead);
				Console.WriteLine("Synchronous Read");
				Console.WriteLine("\tCount of bytes: {0}", bytesRead);
				Console.WriteLine("\tData: {0}\n", data);
			}
		}

		private Stream OpenFileAsynchronously(string aFileName)
		{
			FileStream stream = 
				new FileStream(aFileName, FileMode.Open,
				FileAccess.Read, FileShare.Read, READ_BUF_SIZE,
				true);
			return stream;
		}

		public void AsynchronousEndReadFile()
		{
			Stream stream = OpenFileAsynchronously(FILE_NAME);
			using (stream)
			{
				byte[] buf = new byte[READ_BUF_SIZE];
				IAsyncResult readResult = stream.BeginRead(
					buf, 0, buf.Length, null, null);
				int bytesRead = stream.EndRead(readResult);
				string data = Encoding.ASCII.GetString(buf, 0, bytesRead);
				Console.WriteLine("Asynchronous End Read");
				Console.WriteLine("\tCount of bytes: {0}", bytesRead);
				Console.WriteLine("\tData: {0}\n", data);
			}
		}

		void OnReadDone(IAsyncResult aResult)
		{
			FileReadState state = (FileReadState) aResult.AsyncState;
			Stream stream = state.Stream;
			using (stream)
			{
				int bytesRead = stream.EndRead(aResult);
				byte[] buf = state.Buffer;
				string data = Encoding.ASCII.GetString(buf, 0, bytesRead);
				Console.WriteLine("Asynchronous Callback Read");
				Console.WriteLine("\tCount of bytes: {0}", bytesRead);
				Console.WriteLine("\tData: {0}\n", data);
			}
		}

		public void AsynchronousCallbackReadFile()
		{
			Stream stream = OpenFileAsynchronously(FILE_NAME);
			byte[] buf = new byte[READ_BUF_SIZE];
			FileReadState state = new FileReadState(stream, buf);
			AsyncCallback readDone = new AsyncCallback(this.OnReadDone);
			IAsyncResult readResult = stream.BeginRead(
				buf, 0, buf.Length, readDone, state);
		}
		
		public void AsynchronousPollReadFile()
		{
			Stream stream = OpenFileAsynchronously(FILE_NAME);
			byte[] buf = new byte[READ_BUF_SIZE];
			IAsyncResult readResult = stream.BeginRead(
				buf, 0, buf.Length, null, null);

			Console.Write("Asynchronous Poll Read");
			while (! readResult.IsCompleted)
			{
				Thread.Sleep(WAIT_TIMEOUT);
				Console.Write(".");
			}
			Console.WriteLine();

			using (stream)
			{
				int bytesRead = stream.EndRead(readResult);
				string data = Encoding.ASCII.GetString(buf, 0, bytesRead);
				Console.WriteLine("\tCount of bytes: {0}", bytesRead);
				Console.WriteLine("\tData: {0}\n", data);
			}
		}

		public void AsynchronousWaitReadFile()
		{
			Stream stream = OpenFileAsynchronously(FILE_NAME);
			byte[] buf = new byte[READ_BUF_SIZE];
			IAsyncResult readResult = stream.BeginRead(
				buf, 0, buf.Length, null, null);

			Console.Write("Asynchronous Wait Read");
			bool finished;
			do
			{
				finished = readResult.AsyncWaitHandle.
					WaitOne(WAIT_TIMEOUT, false);
				Console.Write(".");
			}
			while (! finished);
			Console.WriteLine();

			using (stream)
			{
				int bytesRead = stream.EndRead(readResult);
				string data = Encoding.ASCII.GetString(buf, 0, bytesRead);
				Console.WriteLine("\tCount of bytes: {0}", bytesRead);
				Console.WriteLine("\tData: {0}\n", data);
			}
		}
	}

	class FileReadState
	{
		private Stream mStream;
		private byte[] mBuffer;

		public Stream Stream
		{
			get
			{
				return mStream;
			}
		}

		public byte[] Buffer
		{
			get
			{
				return mBuffer;
			}
		}

		public FileReadState(Stream aStream, byte[] aBuffer)
		{
			mStream = aStream;
			mBuffer = aBuffer;
		}
	}

	class AsyncFileReadDemo
	{

		static void Main(string[] args)
		{
			FileReaderDemo demo = new FileReaderDemo();
			demo.CreateFile();
			demo.SynchronousReadFile();
			demo.AsynchronousEndReadFile();
			demo.AsynchronousCallbackReadFile();
			Thread.Sleep(500); // Wait for the previous call get finished
			demo.AsynchronousPollReadFile();
			demo.AsynchronousWaitReadFile();
		}
	}
}